/*
 * String funcs.
 * Strings that contain HobbiRewl controlseqs should be manipulated with
 * these.
 */

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <string.h>

#include "hr_defines.h"
#include "hr_fonts.h"


/* Returns number of "items" in a string.*/
int hr_strlen(unsigned char *string)
{
	int c,len=0,formatmode=0;
	if(string!=NULL){
		for(c=0;c<strlen(string);c++){
			switch(*(string+c)){
			 case FONT_TOGGLECHAR:
				break;
			 case FONT_FORMATCHAR:
				formatmode^=1;
				break;
			 default:
				if(!formatmode)
				  len++;
				break;
			}
		}
	}
	return len;
}


/* Returns
 * 1-255 for a character
 * 0     for End-of-string or error
 * or STRING_TOGGLESEQ
 * or STRING_FORMATSEQ
 */
int hr_string_item(unsigned char *string,int position)
{
	int c,curpos=0;
	int formatmode=0;
	if(string!=NULL){
		if(position<hr_strlen(string)){
			for(c=0;c<strlen(string);c++){
				switch(*(string+c)){
				 case FONT_TOGGLECHAR:
					if(curpos==position){
						/* Check whether the FONT_TOGGLECHAR pair is printable.*/
						switch(*(string+c+1)){
						 case 't':
							return FONT_TOGGLECHAR;
							break;
						 case 'f':
							return FONT_FORMATCHAR;
							break;
						 default:
							return STRING_TOGGLESEQ;
							break;
						}
					}
					++curpos;
					++c;
					break;
				 case FONT_FORMATCHAR:
					if(!formatmode){
						if(curpos==position)
						  return STRING_FORMATSEQ;
						++curpos;
					}
					formatmode^=1;
					break;
				 default:
					if(!formatmode){
						if(curpos==position)
						  return *(string+c);
						++curpos;
					}
					break;
				}
			}
		}
	}
	return 0;
}


/* Returns offset of a string item within a string
 * in bytes. -1 == error.*/
int hr_string_item_offset(unsigned char *text,int item)
{
	int c,curpos=0;
	int formatmode=0;
	if(text!=NULL){
		/* Are there that many items in string?*/
		if(item<hr_strlen(text)){
			/* Scan every byte.*/
			for(c=0;c<strlen(text);c++){
				switch(*(text+c)){
				 case FONT_TOGGLECHAR:
					if(curpos==item)
					  return c;
					++curpos;
					++c;
					break;
				 case FONT_FORMATCHAR:
					if(!formatmode){
						if(curpos==item)
						  return c;
						++curpos;
					}
					formatmode^=1;
					break;
				 default:
					if(!formatmode){
						if(curpos==item)
						  return c;
						++curpos;
					}
					break;
				}
			}
		}
	}
	return -1;
}


/* Modifies given flags as described in *string.*/
int hr_parse_controlseq(unsigned char *string,int mflags,int flags)
{
	unsigned char c=0;
	if(string!=NULL){
		switch(*string){
		 case FONT_TOGGLECHAR:
			/* Scanning direction */
			if(flags & STRING_RIGHT)
			  c=*(++string);
			else if(flags & STRING_LEFT)
			  c=*(--string);
			switch(c){
			 case 'b':
				mflags ^= FONT_BOLD;
				break;
			 case 'u':
				mflags ^= FONT_UNDERLINED;
				break;
			 case 'r':
				mflags ^= FONT_REVERSE;
				break;
			 case 's':
				mflags ^= FONT_SINGLECOLBGR;
				break;
			 case 'i':
				mflags ^= FONT_ITALICS;
				break;
			 case 'z':
				mflags ^= FONT_SCALED;
				break;
			 case 'x':
				mflags ^= FONT_XMULTIPLIED;
				break;
			 case 'y':
				mflags ^= FONT_YMULTIPLIED;
				break;
			 case 'o':
				mflags ^= FONT_ROTATED;
				break;
			 case 'p':
				mflags ^= FONT_PROPORTIONAL;
				break;
			}
			break;
		 case FONT_FORMATCHAR:
			/* ADD! */
			break;
		}
	}
	return mflags;
}


/* Inserts "string" in some position in "target" string.
 * Returns NULL if an error occurred, "target" otherwise.
 * "target" better have enough space for the new string, or else... */
unsigned char *hr_insert_string(unsigned char *target,int position,
										  unsigned char *string,unsigned char *addition)
{
	int bytes1,o2;
	if((target!=NULL)&&(string!=NULL)&&(addition!=NULL)){
		if(position<=hr_strlen(target)){
			/* Calc number of bytes in the part of the string
			 * left from "position".*/
			if(position>0)
			  bytes1=hr_string_item_offset(string,position);
			else
			  bytes1=0;
			/* Calc offset of first byte in the part of the string
			 * right from "position".*/
			if(position<hr_strlen(string))
			  o2=hr_string_item_offset(string,position);
			else
			  o2=0;
			/* Build new string in target */
			*target=0;
			if(bytes1>0)
			  strncat(target,string,bytes1);
			strcat(target,addition);
			if(o2>0)
			  strcat(target,string+o2);
			return target;
		}
	}
	return NULL;
}


/* Deletes string item, be it a character or a controlseq.*/
void hr_delete_string_item(unsigned char *string,int position)
{
	int o1,o2;
	if(string!=NULL){
		if(position<hr_strlen(string)){
			/* Calc offsets of previous/next string item.*/
			if(position>0)
			  o1=hr_string_item_offset(string,position);
			else
			  o1=0;
			if(position<hr_strlen(string)-1)
			  o2=hr_string_item_offset(string,position+1);
			else
			  o2=0;
			/* Now either insert a zero to o1, or memmove o2 to o1+1.*/
			if(o2==0)
			  *(string+o1)=0;
			else
			  memmove(string+o1,string+o2,strlen(string+o2));
		}
	}
}


/* Returns visible number of chars in string containing formatting codes.
 * For example, length of "Guggu~BBold~b." would be 10. */
int hr_visible_strlen(unsigned char *string)
{
	int len=0,c;
	if(string!=NULL){
		for(c=0;c<hr_strlen(string);c++){
			if(hr_string_item(string,c)>0)
			  ++len;
		}
	}
	return len;
}


/* Returns width of a string, or part of a string, in pixels.*/
int hr_string_xpixels(unsigned char *string,struct hr_font *font,
							 int offset,int len,int flags)
{
	/* El Slowo! */
	return hr_write(0,0,0,0,string,font,offset,len,flags|FONT_INVISIBLE);
}

/* Returns height of a string, or part of a string, in pixels.*/
/* FIX! */
int hr_string_ypixels(unsigned char *string,struct hr_font *font,
							 int offset,int len,int flags)
{
	/* El Slowo! */
	return hr_write(0,0,0,0,string,font,offset,len,
						 flags|FONT_INVISIBLE|FONT_RETURN_HEIGHT);
}


/* Tells how many characters in string fit in n pixels.*/
/* FIX! Check FONT_PROPORTIONAL and size/font changes etc... */
int hr_string_xchars(unsigned char *string,struct hr_font *font,int n,
							int flags)
{
	return (n*(font->width))/hr_visible_strlen(string);
}


/* Returns amount of words in a string */
int hr_words(unsigned char *string)
{
	int c,l,words=0,word=0;
	if(string!=NULL){
		for(l=0;l<hr_strlen(string);l++){
			c=hr_string_item(string,l);
			/* Is this item printable? */
			if(c>0){
				/* If current characters is not a char/digit,
				 * toggle on "word".*/
				if(isalnum(c)){
					if(!word){
						++words;
						word=1;
					}
				}
				else
				  word=0;
			}
			else
			  word=0;
		}
	}
	return words;
}


/* Returns pointer to a word within a string.*/
unsigned char *hr_word(unsigned char *string,int xword)
{
	int c,l,words=0,word=0;
	if(string!=NULL){
		for(l=0;l<hr_strlen(string);l++){
			c=hr_string_item(string,l);
			if(c>0){
				/* If current characters is not a char/digit, add amount of words.*/
				if(isalnum(c)){
					if(!word){
						if(++words == xword)
						  return (string+hr_string_item_offset(string,l));
						word=1;
					}
				}
				else
				  word=0;
			}
			else
			  word=0;
		}
	}
	return NULL;
}
